/*
  Todd Goldfinger

  Library for reading fixed-field file headers.

  You can use this under GPL 2.0.
*/
#include "_dumphead.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>

#ifdef WORD_BIT
#    if WORD_BIT!=32
#      error Need 4 byte integers
#    endif 
#else
#    warning "***don't know int size, building anayway - may not work***"
#endif

#define TOSTR(MACSTR) _FSTR(MACSTR)
#define _FSTR(MACSTR) "%" #MACSTR "s%15s%" #MACSTR "hd ?BIG %s"

static unsigned int getNbytesBig(unsigned char *d, int nbytes);
static unsigned int getNbytesLittle(unsigned char *d, int nbytes);

int getHead(struct header **head, const char *file) 
{
#define HEAD (*head)
#define QUIT(err) _QUIT(err)
#define _QUIT(err)           \
{                            \
    if(fin) fclose(fin);     \
    if(fdesc) fclose(fdesc); \
    free(buf);               \
    freeHead(*head);         \
    return err;              \
}
    FILE *fdesc=NULL,*fin=NULL;
    char *c,type[16],scratch[STRSIZE];
    /* Scratch is first used to store a file name.
       Then it's used to hold the endian field (if there is one) */
    unsigned char *buf=NULL;
    int max=4,i=0,n;
    unsigned int (*getbytes)(unsigned char *,int)=getNbytesLittle/*assume little*/;
    void *tmp=NULL;

    if(!(*head=malloc(sizeof(struct header)))) QUIT(NOMEMORY);
    HEAD->nflds = 0;
    if(!(HEAD->fields=malloc(sizeof(struct field)*max))) QUIT(NOMEMORY)
    if(!(fin=fopen(file,"rb"))) QUIT(OPENINERROR);
    if(!(c=strrchr(file,'.'))) QUIT(UNKOWNTYPE);
    snprintf(scratch,STRSIZE,"%s/%s",FTYPEDIR,c+1);
    //sprintf(scratch,"/Users/class02/goldfita/dumphead/%s",c+1);
    if(!(fdesc=fopen(scratch,"r"))) QUIT(OPENDESCERROR);
    
    while((n=fscanf(fdesc,TOSTR(STRSIZE),HEAD->fields[i].name,
                    type,&HEAD->fields[i].size,scratch)) >= 3) {
        if(HEAD->fields[i].size==0) {
            getbytes = getNbytesBig;
            continue;
        }
        
        if(!(HEAD->fields[i].val=malloc(HEAD->fields[i].size+1))) 
            QUIT(NOMEMORY);
	HEAD->nflds++;
	if(!(buf=malloc(HEAD->fields[i].size))) 
	    QUIT(NOMEMORY);
        if(fread(buf,sizeof(char),HEAD->fields[i].size,fin)!=HEAD->fields[i].size)
            QUIT(READINERROR);
        
        if(!strcmp(type,"int")) {
            HEAD->fields[i].type = UINTEGER;
            if(HEAD->fields[i].size>4)
                QUIT(INTTOOBIG);
            *(unsigned int *)HEAD->fields[i].val = getbytes(buf,HEAD->fields[i].size);
        } else if(!strcmp(type,"string")) {
            memcpy(HEAD->fields[i].val,buf,HEAD->fields[i].size);
            ((char *)HEAD->fields[i].val)[HEAD->fields[i].size] = '\0';
            HEAD->fields[i].type = STRING;
            if(n==4 && !strcmp(scratch,HEAD->fields[i].val))
                getbytes = getNbytesBig;
        } else if(!strcmp(type,"other")) {
            memcpy(HEAD->fields[i].val,buf,HEAD->fields[i].size);
            HEAD->fields[i].type = OTHER;
        } else
            QUIT(UNKNOWNTYPE);
        
        if(++i>=max) {
            max <<= 1;
            if((tmp=realloc(HEAD->fields,max*sizeof(struct field))))
                HEAD->fields = tmp;
            else
                QUIT(NOMEMORY);
        }
        free(buf);
    }
    
    return 0;
}


void freeHead(struct header *head)
{
    int i;

    if(head && head->fields)
	for(i=0; i<head->nflds; i++)
	    if(head->fields[i].val)
		free(head->fields[i].val);
    free(head->fields);
}

static unsigned int getNbytesBig(unsigned char *d, int nbytes)
{
    unsigned int b = (unsigned int)d[0];

    if(nbytes>1)
	b = (b<<8)|((unsigned int)d[1]);
    if(nbytes>2)
	b = (b<<8)|((unsigned int)d[2]);
    if(nbytes>3)
	b = (b<<8)|((unsigned int)d[3]);
    return b;
}

static unsigned int getNbytesLittle(unsigned char *d, int nbytes)
{
    unsigned int b = (unsigned int)d[0];
    
    if(nbytes>1)
	b |= (((unsigned int)d[1])<<8);
    if(nbytes>2)
	b |= (((unsigned int)d[2])<<16);
    if(nbytes>3)
	b |= (((unsigned int)d[3])<<24);
    return b;
}

/*
  It would be a little work, but it might be possible to extend this
  to work with variable length headers or headers with conditional
  bytes.  You could also extend it to work with all the meta data in a
  file instead of just the header.  It would require a simple language
  instead of the very simple file format I designed.  For example,
  after the size field, you might have some type of conditional like
  fmt? (for a wave file).

  RiffType	char	4	fmt? goto(xxx)

  It's not quite that simple, but that's the idea.
*/
