/*
  Todd Goldfinger

  Tool for displaying fixed-field file headers.
  
  -f filename
  -h hex mode
  -s print byte sizes

  You can use this under GPL 2.0.
*/
#include "_dumphead.h"

#include <unistd.h>
#include <dlfcn.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

void options(int argc, char *argv[], int *hex, int *size, const char **fname);

int main(int argc, char *argv[])
{
    struct header *head;
    int res,i,j,hex=0,size=0;
    const char *file=NULL,*err;
    char formatstr[256],formatstr2[256];
    void* lib_handle;
    int (*getHead)(struct header **, const char *file);
    int (*freeHead)(struct header *);

    options(argc,argv,&hex,&size,&file);
    if(!file)
	return -1;
   
    lib_handle = dlopen(DUMPHEAD_SO,RTLD_LAZY);
    if(!lib_handle) {
        fprintf(stderr,"Error during dlopen(): %s\n",dlerror());
        exit(-1);
    }
   
    getHead = dlsym(lib_handle,"getHead");
    err = dlerror();
    if(err) {
        fprintf(stderr,"Error locating 'getHead' - %s\n",err);
        exit(-1);
    }
    freeHead = dlsym(lib_handle,"freeHead");
    err = dlerror();
    if(err) {
        fprintf(stderr,"Error locating 'freeHead' - %s\n",err);
        exit(-1);
    }
    
    if((res=(*getHead)(&head,file))) {
        fprintf(stderr,"error %d\n",res);
        dlclose(lib_handle);
        return -1;
    }

    if(size) {
        sprintf(formatstr,"%%-32s %%16%s %%8d\n",hex?"x": "u");
        strcpy(formatstr2,"%-32s %16s %8d\n");
    } else {
        sprintf(formatstr,"%%-32s %%16%s\n",hex?"x": "u");
        strcpy(formatstr2,"%-32s %16s\n");
    }
    for(i=0; i<head->nflds; i++)
        if(head->fields[i].type==UINTEGER)
            printf(formatstr,
                   head->fields[i].name,
                   *(unsigned int *)head->fields[i].val,
                   head->fields[i].size);
        else if(head->fields[i].type==STRING)
            printf(formatstr2,
                   head->fields[i].name,
                   (char *)head->fields[i].val,
                   head->fields[i].size);
    	else {
            printf("%-32s ",head->fields[i].name);
            if(hex)
                for(j=0; j<head->fields[i].size; j++)
                    printf("%x",((char *)(head->fields[i].val))[j]);
            else
                printf("%16s","");
            if(size)
                printf(" %8d\n",head->fields[i].size);
            else
                puts("\n");
        }
    (*freeHead)(head);
    
    dlclose(lib_handle);
    return 0;
}

void options(int argc, char *argv[], int *hex, int *size, const char **fname)
{
    int c;
    
    while((c=getopt(argc,argv,"hsf:")) != -1) {
        switch(c) {
        case 'h':
            *hex = 1;
            break;
        case 's':
            *size = 1;
            break;
        case 'f':
            *fname = optarg;
            break;
        case '?':
	    exit(-1);
        }
    }
}

