#include "Python.h"
#include "../_dumphead.h"

#include <dlfcn.h>
#include <stdio.h>

//python setup.py build_ext --inplace

#define DH_ERROR_I(err,buf) {              \
    sprintf(buf,"error number==%d",err);   \
    PyErr_SetString(ReadHeaderFailed,buf); \
    return NULL;                           \
}

#define DH_ERROR_S(err,name,buf) {                           \
    snprintf(buf,256,"Trouble with '%s': %s",name,err);      \
    PyErr_SetString(ReadHeaderFailed,buf);                   \
    return NULL;                                             \
}

static PyObject *buildDict(struct header *head, char *file);
static PyObject *_getHead(PyObject *self, PyObject *args);
PyMODINIT_FUNC initdumphead(void);

static PyObject *ReadHeaderFailed;

static PyMethodDef dumphead_methods[] = {
    {"getHead",_getHead,METH_VARARGS,"Get header as a dictionary."},
    {NULL,NULL,0,NULL}
};

PyMODINIT_FUNC initpy_dumphead(void) 
{
    PyObject *m;

    m = Py_InitModule("py_dumphead",dumphead_methods);
    ReadHeaderFailed = PyErr_NewException("ReadHeaderFailed.error",NULL,NULL);
    Py_INCREF(ReadHeaderFailed);
    PyModule_AddObject(m,"error",ReadHeaderFailed);
}

static PyObject *_getHead(PyObject *self, PyObject *args)
{
    struct header *head;
    void* lib_handle;
    int (*getHead)(struct header **, const char *file);
    int (*freeHead)(struct header *);
    char buf[256],*err,*file;
    int res;
    PyObject *dict;

    if(!PyArg_ParseTuple(args,"s",&file))
        return NULL;

    lib_handle = dlopen("/usr/local/lib/_dumphead.so",RTLD_LAZY);
    if(!lib_handle)
        DH_ERROR_S(dlerror(),"dlopen()",buf);
    getHead = dlsym(lib_handle,"getHead");
    err = dlerror();
    if(err)
        DH_ERROR_S(err,"getHead",buf);
    freeHead = dlsym(lib_handle,"freeHead");
    err = dlerror();
    if(err)
        DH_ERROR_S(err,"freeHead",buf);

    if((res=(*getHead)(&head,file))) {
        DH_ERROR_I(res,buf);
        dlclose(lib_handle);
    }
    dict = buildDict(head,file);
    (*freeHead)(head);
    dlclose(lib_handle);
    return dict;
}

static PyObject *buildDict(struct header *head, char *file)
{
    int i,res;
    PyObject *dict,*name,*attr;
    
    if(head->nflds<1) {
        Py_INCREF(Py_None);
        return Py_None;
    } else {
        if(head->fields[0].type==UINTEGER)
            dict = Py_BuildValue("{s:i}",head->fields[0].name,
                                 *(unsigned int *)head->fields[0].val);
        else if(head->fields[0].type==STRING)
            dict = Py_BuildValue("{s:s}",head->fields[0].name,
                                 (char *)head->fields[0].val);
        else {
            dict = Py_BuildValue("{s:s}",head->fields[0].name,
                                 (char *)head->fields[0].val);
        }
    }

    for(i=1; i<head->nflds; i++) {
        name = PyString_FromString(head->fields[i].name);//new ref
        if(head->fields[i].type==UINTEGER)
            attr = PyInt_FromLong(*(unsigned int *)head->fields[i].val);//new ref
        else
            attr = PyString_FromString((char *)head->fields[i].val);//new ref
        
        if(name&&attr) {
            res=PyDict_SetItem(dict,name,attr);
            Py_DECREF(name);
            Py_DECREF(attr);
            if(res)
                return NULL;
        } else 
            return NULL;
    }

    return dict;
}

