#include <stddef.h>

#include "ruby.h"
#include "../_dumphead.h"

#include <dlfcn.h>

//ruby extconf.rb
//make

//http://www.ruby-doc.org/
//http://www.rubycentral.com/book
//http://www.tarkblog.org/rubyextension/part-2/
//tar cvzf dhead.tgz dumphead

static VALUE populate_hash(struct header *head);

VALUE dumphead;

VALUE dhead_get(VALUE class, VALUE file)
{
    VALUE h=Qnil;
    struct header *head;
    int res;
    void *lib_handle;
    int (*getHead)(struct header **, const char *file);
    int (*freeHead)(struct header *);
    const char *err, *fname;
    

    fname = StringValuePtr(file);
    lib_handle = dlopen("/usr/local/lib/_dumphead.so",RTLD_LAZY);
    if(!lib_handle) {
	dlclose(lib_handle);
        rb_raise(rb_eRuntimeError,"dlopen failed - %s",dlerror());
    }
   
    getHead = dlsym(lib_handle,"getHead");
    err = dlerror();
    if(err) {
	dlclose(lib_handle);
        rb_raise(rb_eRuntimeError,"Error locating getHead - %s",dlerror());
    }
    freeHead = dlsym(lib_handle,"freeHead");
    err = dlerror();
    if(err) {
	dlclose(lib_handle);
        rb_raise(rb_eRuntimeError,"Error locating freeHead - %s",dlerror());
    }
    
    if((res=(*getHead)(&head,fname))) {
	dlclose(lib_handle);
        rb_raise(rb_eRuntimeError,"Error code is - %d",res);
    } else {
	h = populate_hash(head);
	(*freeHead)(head);
    }
    dlclose(lib_handle);
    return h;
}


void Init_rb_dumphead() 
{
    dumphead = rb_define_class("DumpHead",rb_cObject);
    rb_define_singleton_method(dumphead,"Head",dhead_get,1);
}

static VALUE populate_hash(struct header *head)
{
    int i;
    VALUE key,val,h = rb_hash_new();
    
    for(i=0; i<head->nflds; i++) {
	key = rb_str_new2(head->fields[i].name);
	if(head->fields[i].type==UINTEGER)
	    val = INT2FIX(*(unsigned int *)head->fields[i].val);
	else
	    val = rb_str_new2((char *)head->fields[i].val);
	rb_hash_aset(h,key,val);
    }
    return h;
}
